/*
  Copyright (c) 2004,2005. PathScale, Inc. All rights reserved.
  Copyright (c) 2015, Intel Corporation

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of Intel Corporation nor the names of its contributors
      may be used to endorse or promote products derived from this software
      without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.    

*/

/*
 * Derived from code supplied by Ohio State University:
 * Copyright (C) 2002-2003 the Network-Based Computing Laboratory
 * (NBCL), The Ohio State University.  
 *  http://nowlab.cis.ohio-state.edu/projects/mpi-iba/
 */

#include "mpi.h"
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <unistd.h>
#include <getopt.h>
#include <string.h>
#include <errno.h>

#define MYBUFSIZE 16*1024
#define MAX_REQ_NUM 1000

char s_buf[MYBUFSIZE];
char r_buf[MYBUFSIZE];
MPI_Request request[MAX_REQ_NUM];
MPI_Status  reqstat[MAX_REQ_NUM];
int window_size = 64;

void usage (void) {
     extern char *__progname;
     printf ("Usage: mpirun -np <procs> %s [[-q] [-c] [-m <message size>]\n",__progname);
     printf ("       [-l <repetitions for latency test] [-h]]\n");
     printf ("    -q   - Run in quiet mode. In quiet mode only progress\n");
     printf ("           indication is printed out and the results are\n");
     printf ("           printed at the end of the run.\n");
     printf ("           A second 'q' will suppress the printing of latency\n");
     printf ("           and bandwidth results all together.\n");
     printf ("    -c   - For each node pair, only one direction is tested, \n");
     printf ("           resulting in N*(N-1)/2 tests.  Each node will be \n");
     printf ("           tested as sender and receiver alternately. This is \n");
     printf ("           a 'c'heckerboard pattern of coverage on the NxN \n");
     printf ("           node-to-node matrix.  This option is useful to \n");
     printf ("           reduce run time on large clusters.  \n");
     printf ("    -l <repetitions for latency test> \n");
     printf ("         - Specify repetitions for latency test. \n");
     printf ("           The default is 1000.\n");
     printf ("    -m <message size> \n");
     printf ("         - Message size can be specified other than the \n");
     printf ("           default 16KB for bandwidth test. The input will \n");
     printf ("           be convert to the nearest power of two.\n"); 
     printf ("    -h   - Show this help screen.\n");
     return;
}

int near_power_2 (int input);

int main(int argc, char *argv[])
{
     
    int myid, numprocs, i;
    int size = MYBUFSIZE;
    MPI_Status stat;

    double t_start = 0.0, t_end = 0.0;
    int latloop = 1000, latskip = 10;
    int bwloop = 10, bwskip = 2;
    int src, dest, quiet=0, uni=0, size_raw, a, b, bit = 0, b_ini, j;
    double **results_lat = NULL, **results_bw = NULL;
    
    MPI_Init(&argc, &argv);
    MPI_Comm_size(MPI_COMM_WORLD, &numprocs);
    MPI_Comm_rank(MPI_COMM_WORLD, &myid);

    while ((i=getopt (argc, argv, "hqcm:l:")) != -1) {
	 switch (i) {
         case 'l':
              latloop= atoi(optarg);
              break;
	 case 'q':
	      quiet += 1;
	      break;
         case 'c':
              uni += 1; 
              break;
         case 'm':
              size_raw = atoi(optarg);
              size = near_power_2 (size_raw);
              break;
	 case 'h':
	      if (myid == 0) {
		   usage ();
	      }
              MPI_Finalize();
              return 0;
	      break;
	 default:
	      if (myid == 0) {
		   usage ();
		   MPI_Abort (MPI_COMM_WORLD, 0);
	      }
	      break;
	 }
    }

    if (quiet > 2) quiet = 2;
    
    if (myid == 0 && quiet > 0) {
	 /* allocate the results array based on how many
	    node we have */
	 results_lat = calloc (numprocs, sizeof (double *));
	 results_bw = calloc (numprocs, sizeof (double *));
	 for (src = 0; src < numprocs; src++) {
	      results_lat[src] = calloc (numprocs, sizeof (double));
	      results_bw[src] = calloc (numprocs, sizeof (double));
	 }
    }
    /*touch the data*/
    for (i=0; i<size; i++ ){
         s_buf[i]='a';
         r_buf[i]='b';
    }

    b_ini = 0; 
    for (a = 0; a < numprocs; a++) {
        if (uni == 1) {
             b_ini = a + 1;
             bit = 1 ;
        }
        for (b = b_ini ; b < numprocs; b++) {
            if (a == b)
                continue;
            src = a;
            dest = b;
            if (uni == 1) { 
               if (bit % 2 == 0) { 
                    src = b; 
                    dest = a;
               }
                bit ++;
            }
                
            if (myid == src) {
                for (i = 0; i < latloop + latskip; i++) {
                    if (i == latskip)
                        t_start = MPI_Wtime();
                    MPI_Send((void*)0, 0, MPI_CHAR, dest, 1, MPI_COMM_WORLD);
                    MPI_Recv((void*)0, 0, MPI_CHAR, dest, 1, MPI_COMM_WORLD,
                            &stat);
                }
                t_end = MPI_Wtime();

            } else if (myid == dest) {
                for (i = 0; i < latloop + latskip; i++) {
                    MPI_Recv((void*)0, 0, MPI_CHAR, src, 1, MPI_COMM_WORLD,
                            &stat);
                    MPI_Send((void*)0, 0, MPI_CHAR, src, 1, MPI_COMM_WORLD);
                }
            }

            if (myid == src) {
                double latency;
                latency = (t_end - t_start) * 1.0e6 / (2.0 * latloop);
		if (myid == 0)
		     if (quiet > 0) {
			  results_lat[src][dest] = latency;
		     } else {
			  fprintf (stdout, "[%d<->%d]\t\t%0.2fus\t\t", src, dest, latency);
		     }
		else
		     MPI_Send (&latency, 1, MPI_DOUBLE, 0, 1, MPI_COMM_WORLD);
            } else if (myid == 0) {
		 double latency;
		 MPI_Recv (&latency, 1, MPI_DOUBLE, src, 1, MPI_COMM_WORLD, &stat);
		 if (quiet > 0) {
		      results_lat[src][dest] = latency;
		 } else {
		      fprintf (stdout, "[%d<->%d]\t\t%0.2fus\t\t", src, dest, latency);
		 }
	    }
		 

            if (myid == src) {
                for ( i=0; i< bwskip; i++ ) {
                    for ( j=0; j< window_size;j++)
                        MPI_Isend(s_buf, size, MPI_CHAR, dest, i, MPI_COMM_WORLD, request+j);
                    MPI_Waitall(window_size, request,reqstat);
                }
                MPI_Recv(NULL, 0, MPI_CHAR, dest, i+1000, MPI_COMM_WORLD,&reqstat[0]);

                t_start=MPI_Wtime();
                for ( i=0; i< bwloop; i++ ) {
                    for ( j=0; j< window_size;j++)
                        MPI_Isend(s_buf, size, MPI_CHAR, dest, i, MPI_COMM_WORLD, request+j);
                    MPI_Waitall(window_size, request,reqstat);
                }
                MPI_Recv(NULL, 0, MPI_CHAR, dest, i+1000, MPI_COMM_WORLD, &reqstat[0]);
                t_end=MPI_Wtime();

            } else if (myid == dest) {
                for ( i=0; i< bwskip; i++ ) {
                    for ( j=0; j< window_size;j++)
                        MPI_Irecv(r_buf, size, MPI_CHAR, src,i,MPI_COMM_WORLD,request+j);
                    MPI_Waitall(window_size, request,reqstat);
                }
                MPI_Send(NULL, 0, MPI_CHAR, src, i + 1000, MPI_COMM_WORLD);
                for ( i=0; i< bwloop; i++ ) {
                    for ( j=0; j< window_size;j++)
                        MPI_Irecv(r_buf, size, MPI_CHAR, src,i,MPI_COMM_WORLD,request+j);
                    MPI_Waitall(window_size, request,reqstat);
                }
                MPI_Send(NULL, 0, MPI_CHAR, src, i + 1000, MPI_COMM_WORLD);
            }

            if ( myid == src ) {
                double bw;
                bw = ((double)size*(double)bwloop)*(double)window_size/((t_end-t_start)*(1.0e6));
                if (myid == 0)
		     if (quiet > 0) {
			  results_bw[src][dest] = bw;
		     } else {
			  fprintf (stdout, "%f (MillionBytes/sec)\n", bw);
		     }
		else
		     MPI_Send (&bw, 1, MPI_DOUBLE, 0, 1, MPI_COMM_WORLD);
	    } else if (myid == 0) {
		 double bw;
		 MPI_Recv (&bw, 1, MPI_DOUBLE, src, 1, MPI_COMM_WORLD, &stat);
		 if (quiet > 0) {
		      results_bw[src][dest] = bw;
		 } else {
		      fprintf (stdout, "%f (MillionBytes/sec)\n", bw);
		 }
		 if (quiet > 0) {
		      printf (".");
		      fflush (stdout);
		 }
            }
        }
        MPI_Barrier(MPI_COMM_WORLD);
    }

    if (myid == 0) {
	 switch (quiet) {
	 case 1:
	      printf ("\n");
              b_ini = 0 ;
	      for (a = 0; a < numprocs; a++) {
                   if (uni == 1) {
                        b_ini = a + 1 ;
                        bit = 1 ;
                   }
                   for (b = b_ini ; b < numprocs; b++) {
                        if ( a == b) continue;
                        src = a;
                        dest = b;
                        if (uni == 1) { 
                             if (bit % 2  == 0) {
                                  src = b;
                                  dest = a;
                             }
                             bit ++ ;
                        }
			fprintf (stdout, "[%d<->%d]\t\t%0.2fus\t\t%f (MillionBytes/sec)\n", src, dest, results_lat[src][dest], results_bw[src][dest]);
                  }
              }
	      break;
	 case 2:
	      printf ("\n");
	      break;
	 }
    }

    MPI_Finalize();
    return 0;
}

int near_power_2 (int input)
{	
     int low, high = 1;
     while ( high < input ) {
           high <<= 1;
     } 
     low = high >> 1; 
     if (high-input > input-low) 
          return low;
     else
          return high;
}
