/*
 * Copyright (c) 2013-2018 Intel, Inc. All rights reserved
 *
 * Copyright (c) 2015 Cisco Systems, Inc.  All rights reserved.
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */

#ifndef MTL_OFI_TYPES_H_HAS_BEEN_INCLUDED
#define MTL_OFI_TYPES_H_HAS_BEEN_INCLUDED

#include "mtl_ofi.h"

BEGIN_C_DECLS

/**
 * MTL Module Interface
 */
typedef struct mca_mtl_ofi_module_t {
    mca_mtl_base_module_t base;

    /** Fabric Domain handle */
    struct fid_fabric *fabric;

    /** Access Domain handle */
    struct fid_domain *domain;

    /** Address vector handle */
    struct fid_av *av;

    /** Completion queue handle */
    struct fid_cq *cq;

    /** Endpoint to communicate on */
    struct fid_ep *ep;

    /** Endpoint name length */
    size_t epnamelen;

    /** "Any source" address */
    fi_addr_t any_addr;

    /** Optional user-specified OFI provider name */
    char *provider_name;

    /** Maximum inject size */
    size_t max_inject_size;

    /** Maximum number of CQ events to read in OFI Progress */
    int ofi_progress_event_count;

    /** CQ event storage */
    struct fi_cq_tagged_entry *progress_entries;

    /** Use FI_REMOTE_CQ_DATA*/
    bool fi_cq_data;

    /** MPI tag bits in OFI Tag */
    unsigned long long mpi_tag_mask;

    /** Synchronous protocol tag bits */
    unsigned long long sync_send;
    unsigned long long sync_send_ack;
    unsigned long long sync_proto_mask;

} mca_mtl_ofi_module_t;

extern mca_mtl_ofi_module_t ompi_mtl_ofi;

typedef struct mca_mtl_ofi_component_t {
    /** Base MTL component */
    mca_mtl_base_component_2_0_0_t super;
} mca_mtl_ofi_component_t;

/*OFI TAG:
 * Define 3 different OFI tag distributions:
 * 1) Support FI_REMOTE_CQ_DATA: No need for source rank in the tag
 * 2) DEFAULT_1: Default when no FI_REMOTE_CQ_DATA is supported
 * 3) DEFAULT_2: Alternative default when no FI_REMOTE_CQ_DATA is supported
 *    with more bits for the MPI TAG. Select at build time with
 *    --enable-mtl-ofi-alternative-tag
*/

/* Support FI_REMOTE_CQ_DATA, send the source rank in the CQ data (4 Bytes is the minimum)
 *  01234567 01234567 01234567 012345  67  01234567 01234567 01234567 01234567
 *                                   |    |
 *           context_id              |prot|          message tag
 */
#define MTL_OFI_PROTO_BIT_COUNT         (2)
#define MTL_OFI_HIGHEST_TAG_BIT         (0x8000000000000000ULL)
#define MTL_OFI_MINIMUM_CID_BITS        (8)

#define MTL_OFI_CID_MASK_DATA           (0xFFFFFFFC00000000ULL)
#define MTL_OFI_CID_BIT_COUNT_DATA      (30)
#define MTL_OFI_TAG_MASK_DATA           (0x00000000FFFFFFFFULL)
#define MTL_OFI_TAG_BIT_COUNT_DATA      (32)
#define MTL_OFI_PROTO_HEADER_MASK_DATA  (0x0000000300000000ULL)
#define MTL_OFI_PROTO_MASK_DATA         (0x0000000100000000ULL)
#define MTL_OFI_SYNC_SEND_DATA          (0x0000000100000000ULL)
#define MTL_OFI_SYNC_SEND_ACK_DATA      (0x0000000200000000ULL)

#define MTL_OFI_SET_SEND_BITS_DATA(match_bits, comm_id, tag)         \
    {                                                                \
        match_bits = comm_id;                                        \
        match_bits = (match_bits << (MTL_OFI_TAG_BIT_COUNT_DATA      \
                                     + MTL_OFI_PROTO_BIT_COUNT));    \
        match_bits |= (tag & MTL_OFI_TAG_MASK_DATA);                 \
    }


/* receive posting */
/* Special tags are used for collective operations.
 * MPI_ANY_TAG should not match these special tags.
 * See ompi/mca/coll/base/coll_tags.h
 */
#define MTL_OFI_SET_RECV_BITS_DATA(match_bits, mask_bits, comm_id, tag)   \
    {                                                                     \
        mask_bits  = MTL_OFI_PROTO_MASK_DATA;                             \
                                                                          \
        match_bits = (uint64_t) comm_id;                                  \
        match_bits = (match_bits << (MTL_OFI_PROTO_BIT_COUNT              \
                        +  MTL_OFI_TAG_BIT_COUNT_DATA));                  \
        if (MPI_ANY_TAG == tag) {                                         \
            /* Don't mask the highest TAG bit, special collectives tag */ \
            mask_bits  |= (MTL_OFI_TAG_MASK_DATA>>1);                     \
        } else {                                                          \
            match_bits |= (MTL_OFI_TAG_MASK_DATA & tag);                  \
        }                                                                 \
    }

/*
* DEFAULT_1: Default when no FI_REMOTE_CQ_DATA is supported
*
*  01234567 01234567 01234567 01 234567 01234567 012345  67  01234567 01234567
*                               |                      |    |
*                Comm id        |       source         |prot|  message tag
*/

#define MTL_OFI_CID_MASK_1              (0xFFFFFFC000000000ULL)
#define MTL_OFI_CID_BIT_COUNT_1         (26)
#define MTL_OFI_SOURCE_TAG_MASK_1       (0x0000003FFFFC0000ULL)
#define MTL_OFI_SOURCE_BIT_COUNT_1      (20)
#define MTL_OFI_SOURCE_MASK_1           (0x00000000000000FFULL)
#define MTL_OFI_TAG_MASK_1              (0x000000000000FFFFULL)
#define MTL_OFI_TAG_BIT_COUNT_1         (16)
#define MTL_OFI_PROTO_HEADER_MASK_1     (0x0000000000030000ULL)
#define MTL_OFI_PROTO_MASK_1            (0x0000000000010000ULL)
#define MTL_OFI_SYNC_SEND_1             (0x0000000000010000ULL)
#define MTL_OFI_SYNC_SEND_ACK_1         (0x0000000000020000ULL)

/*
* DEFAULT_2: Alternative default when no FI_REMOTE_CQ_DATA is supported
*
*  01234567 0123 4567 01234567 012345  67  01234567 01234567 01234567 01234567
*               |                    |    |
*    Comm id    |     source         |prot|           message tag
*/

#define MTL_OFI_CID_MASK_2              (0xFFF0000000000000ULL)
#define MTL_OFI_CID_BIT_COUNT_2         (12)
#define MTL_OFI_SOURCE_TAG_MASK_2       (0x000FFFFC00000000ULL)
#define MTL_OFI_SOURCE_BIT_COUNT_2      (18)
#define MTL_OFI_SOURCE_MASK_2           (0x00000000000000FFULL)
#define MTL_OFI_TAG_MASK_2              (0x00000000FFFFFFFFULL)
#define MTL_OFI_TAG_BIT_COUNT_2         (32)
#define MTL_OFI_PROTO_HEADER_MASK_2     (0x0000000300000000ULL)
#define MTL_OFI_PROTO_MASK_2            (0x0000000100000000ULL)
#define MTL_OFI_SYNC_SEND_2             (0x0000000100000000ULL)
#define MTL_OFI_SYNC_SEND_ACK_2         (0x0000000200000000ULL)

#if MTL_OFI_ALTERNATIVE_DEFAULT_TAG

#define MTL_OFI_CID_MASK                MTL_OFI_CID_MASK_2
#define MTL_OFI_CID_BIT_COUNT           MTL_OFI_CID_BIT_COUNT_2
#define MTL_OFI_SOURCE_TAG_MASK         MTL_OFI_SOURCE_TAG_MASK_2
#define MTL_OFI_SOURCE_BIT_COUNT        MTL_OFI_SOURCE_BIT_COUNT_2
#define MTL_OFI_SOURCE_MASK             MTL_OFI_SOURCE_MASK_2
#define MTL_OFI_TAG_MASK                MTL_OFI_TAG_MASK_2
#define MTL_OFI_TAG_BIT_COUNT           MTL_OFI_TAG_BIT_COUNT_2
#define MTL_OFI_PROTO_HEADER_MASK       MTL_OFI_PROTO_HEADER_MASK_2
#define MTL_OFI_PROTO_MASK              MTL_OFI_PROTO_MASK_2
#define MTL_OFI_SYNC_SEND               MTL_OFI_SYNC_SEND_2
#define MTL_OFI_SYNC_SEND_ACK           MTL_OFI_SYNC_SEND_ACK_2

#else

#define MTL_OFI_CID_MASK                MTL_OFI_CID_MASK_1
#define MTL_OFI_CID_BIT_COUNT           MTL_OFI_CID_BIT_COUNT_1
#define MTL_OFI_SOURCE_TAG_MASK         MTL_OFI_SOURCE_TAG_MASK_1
#define MTL_OFI_SOURCE_BIT_COUNT        MTL_OFI_SOURCE_BIT_COUNT_1
#define MTL_OFI_SOURCE_MASK             MTL_OFI_SOURCE_MASK_1
#define MTL_OFI_TAG_MASK                MTL_OFI_TAG_MASK_1
#define MTL_OFI_TAG_BIT_COUNT           MTL_OFI_TAG_BIT_COUNT_1
#define MTL_OFI_PROTO_HEADER_MASK       MTL_OFI_PROTO_HEADER_MASK_1
#define MTL_OFI_PROTO_MASK              MTL_OFI_PROTO_MASK_1
#define MTL_OFI_SYNC_SEND               MTL_OFI_SYNC_SEND_1
#define MTL_OFI_SYNC_SEND_ACK           MTL_OFI_SYNC_SEND_ACK_1

#endif /* #ifdef MTL_OFI_ALTERNATIVE_DEFAULT_TAG */

/* send posting */
#define MTL_OFI_SET_SEND_BITS(match_bits, comm_id, source, tag)      \
    {                                                                \
        match_bits = comm_id;                                        \
        match_bits = (match_bits << MTL_OFI_SOURCE_BIT_COUNT);       \
        match_bits |= (uint64_t)(source & MTL_OFI_SOURCE_MASK);      \
        match_bits = (match_bits << (MTL_OFI_TAG_BIT_COUNT           \
                                     + MTL_OFI_PROTO_BIT_COUNT));    \
        match_bits |= (tag & MTL_OFI_TAG_MASK);                      \
    }

/* receive posting */
/* Special tags are used for collective operations.
 * MPI_ANY_TAG should not match these special tags.
 * See ompi/mca/coll/base/coll_tags.h
 */
#define MTL_OFI_SET_RECV_BITS(match_bits, mask_bits, comm_id, source, tag) \
    {                                                                      \
        match_bits = 0;                                                    \
        mask_bits  = MTL_OFI_PROTO_MASK;                                   \
                                                                           \
        match_bits = comm_id;                                              \
        match_bits = (match_bits << MTL_OFI_SOURCE_BIT_COUNT);             \
                                                                           \
        if (MPI_ANY_SOURCE == source) {                                    \
            match_bits = (match_bits << (MTL_OFI_TAG_BIT_COUNT             \
                                     + MTL_OFI_PROTO_BIT_COUNT));          \
            mask_bits |= MTL_OFI_SOURCE_TAG_MASK;                          \
        } else {                                                           \
            match_bits |= (uint64_t)(source & MTL_OFI_SOURCE_MASK);        \
            match_bits = (match_bits << (MTL_OFI_TAG_BIT_COUNT             \
                                     + MTL_OFI_PROTO_BIT_COUNT));          \
        }                                                                  \
                                                                           \
        if (MPI_ANY_TAG == tag) {                                          \
            /* Don't mask the highes TAG bit, special collectives tag */   \
              mask_bits  |= (MTL_OFI_TAG_MASK>>1);                         \
        } else {                                                           \
            match_bits |= (MTL_OFI_TAG_MASK & tag);                        \
        }                                                                  \
    }

#define MTL_OFI_SET_SYNC_SEND(match_bits)           \
        match_bits |= ompi_mtl_ofi.sync_send

#define MTL_OFI_IS_SYNC_SEND(match_bits)           \
    (ompi_mtl_ofi.sync_send == (ompi_mtl_ofi.sync_proto_mask & match_bits))

#define MTL_OFI_IS_SYNC_SEND_ACK(match_bits)           \
    (ompi_mtl_ofi.sync_send_ack == (ompi_mtl_ofi.sync_proto_mask & match_bits))

#define MTL_OFI_GET_TAG(match_bits)                \
    ((int)(match_bits & ompi_mtl_ofi.mpi_tag_mask))

__opal_attribute_always_inline__ static inline int
mtl_ofi_get_source(struct fi_cq_tagged_entry *wc)
{
    int src;
    if (ompi_mtl_ofi.fi_cq_data) {
        src = (int) wc->data;
    }
    else {
        src = (int)((wc->tag >> (MTL_OFI_PROTO_BIT_COUNT +
                    MTL_OFI_TAG_BIT_COUNT)) & MTL_OFI_SOURCE_MASK);
    }

    return src;
}
END_C_DECLS

#endif /* MTL_OFI_TYPES_H_HAS_BEEN_INCLUDED */
